package lift;

import java.util.Arrays;

enum Direction {
  UP,
  DOWN,
}

/** Lift monitor. Handles lift motion and various state. */
public class LiftMonitor {
  /** the floor the lift is currently on */
  private int floor;
  /** Number of floors */
  private int nfloors;
  /** true if the lift is moving, false if standing still with doors open */
  private boolean moving;
  /** Current direction of the lift */
  private Direction direction;
  /** number of passengers waiting to enter the lift at the various floors */
  private PassengerQueue waitEntry;
  /** number of passengers (in lift) waiting to leave at the various floors */
  private PassengerQueue waitExit;
  /** number of passengers currently in the lift */
  private int load;
  /** view object passed from main */
  private LiftView view;
  /** State (open/closed) of door on current floor */
  private Door door;
  /** Counter of number of passengers currently entering the elevator */
  private int passengersEntering;
  /** Counter of number of passengers currently exiting the elevator */
  private int passengersExiting;

  /** Create new LiftMonitor */
  public LiftMonitor(LiftView v, int floors) {
    floor = 0;
    nfloors = floors;
    moving = true;
    direction = Direction.UP;
    door = new Door();
    waitEntry = new PassengerQueue(nfloors);
    waitExit = new PassengerQueue(nfloors);
    load = 0;
    view = v;
    passengersEntering = 0;
    passengersExiting = 0;
  }

  /** Increases the number of waiting passengers on given floor */
  public synchronized void increaseWaitEntry(int passengerFloor) {
    waitEntry.inc(passengerFloor);
    notifyAll();
  }

  /** Handles passengers waiting to enter arriving lift */
  public synchronized void enterLift(int passengerFloor,
      int passengerDestination) {
    while (floor != passengerFloor || load == 4 || moving ||
        passengersEntering == 4) {
      try {
        wait();
      } catch (InterruptedException e) {
        throw new Error("Monitor.enterLift interrupted " + e);
      }
    }

    waitEntry.dec(floor);
    waitExit.inc(passengerDestination);
    load++;
    passengersEntering++;
    notifyAll();
  }

  /** Called after passengers has completed entry animation */
  public synchronized void enterCompleted() {
    passengersEntering--;
    notifyAll();
  }

  /** Handles passengers waiting to exit lift */
  public synchronized void exitLift(int passengerDestination) {
    while (floor != passengerDestination) {
      try {
        wait();
      } catch (InterruptedException e) {
        throw new Error("Monitor.exitLift interrupted " + e);
      }
    }
    waitExit.dec(passengerDestination);
    load--;
    passengersExiting++;
    notifyAll();
  }

  /** Called after passengers has completed exit animation */
  public synchronized void exitCompleted() {
    passengersExiting--;
    notifyAll();
  }

  /**
   * Handles the conditions when the elevator is to wait for passengers
   * entering/exiting the lift and waiting when there are no passengers waiting
   * on any floor.
   */
  public synchronized int[] liftContinue() {
    while (waitEntry.isEmpty() && waitExit.isEmpty()) {
      try {
        wait();
      } catch (InterruptedException e) {
        throw new Error("LiftContinue no passengers interrupted " + e);
      }
    }

    /** Stops the elevator to allow passengers to enter or exit */
    while (waitEntry.get(floor) > 0 && load != 4 || waitExit.get(floor) > 0 ||
        passengersEntering > 0 || passengersExiting > 0) {
      if (!door.isOpen()) {
        view.openDoors(floor);
        door.open();
      }
      moving = false;
      notifyAll();
      try {
        wait();
      } catch (InterruptedException e) {
        throw new Error("Monitor.liftContinue interrupted " + e);
      }
    }

    if (!moving)
      view.closeDoors();

    if (door.isOpen())
      door.close();

    moving = true;

    calculateDirection();

    /**
     * Sends the elevators current position and next position to the LiftThread
     */
    int[] movingPositions = new int[2];
    movingPositions[0] = floor;
    movingPositions[1] = floor + (direction == Direction.UP ? 1 : -1);
    return movingPositions;
  }

  /** Moves the elevator in the given direction */
  public synchronized void incrementFloor() {
    floor = floor + (direction == Direction.UP ? 1 : -1);
  }

  /**
   * Handles calculation of elevator direction
   * Changes the direction of the elevator if there are no passengers that wants
   * to enter or exit on floors either above or below the elevator
   */
  private void calculateDirection() {
    boolean noPassengersAbove = (waitEntry.totalAbove(floor) == 0 && waitExit.totalAbove(floor) == 0);
    boolean noPassengersBelow = (waitEntry.totalBelow(floor) == 0 && waitExit.totalBelow(floor) == 0);

    if (floor == 0 || noPassengersBelow) {
      direction = Direction.UP; // Always go up if at ground floor or no passengers below
    } else if (floor == nfloors - 1 || noPassengersAbove) {
      direction = Direction.DOWN; // Always go down if at the top floor or no passengers above
    }
  }
}
